/**
 ** A Virtual Host Monitor
 **
 ** Copyright 2009 Gary Wuertz gary@issiweb.com
 **
 ** This program is free software: you can redistribute it and/or modify
 ** it under the terms of the GNU General Public License as published by
 ** the Free Software Foundation, either version 3 of the License, or
 ** (at your option) any later version.
 **
 ** This program is distributed in the hope that it will be useful,
 ** but WITHOUT ANY WARRANTY; without even the implied warranty of
 ** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 ** GNU General Public License for more details.
 **
 ** You should have received a copy of the GNU General Public License
 ** along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include <iostream>
#include <stdexcept>

#include <errno.h>
#include <getopt.h>
#include <signal.h>
#include <stdlib.h>
#include <sys/time.h>
#include <unistd.h>
#include <ncurses.h>

#include "monitor.h"
#include "vhost.h"

#include "stdio.h"

#define  WBUF_LEN 128

using namespace std;
/**
 * Local control structure
 */
static topvhost app;
/**
 * Handle KILL and INT signals
 */
static void catchsig(int s)
{
   app.exit = s;
}
/**
 * Handle window resize signal
 */
static void catchwinch(int s)
{
   app.resize = true;
}
/**
 * Initalize the application - called from monitor completion
 */
void topvhost::begin(int nhosts, int namelen)
{
   this->namelen = namelen;
   this->paint   = true;
   this->total   = nhosts;
   this->sunit = 0;
   for(unsigned i=0;i<this->row.size();i++)
      if (this->row[i].field==' ') {
         string s = this->row[i].format;
         for (unsigned j=0;j<s.length();j++)
            if (s[j]=='\n') this->sunit += 1;
         }
   if (this->sunit==0)
      throw runtime_error("Row requires end of line");
   signal(SIGINT, &catchsig);
   signal(SIGKILL, &catchsig);
   signal(SIGWINCH, &catchwinch);
   WINDOW *win = initscr();         // initialize screen
   this->window  = win;
   keypad(stdscr, TRUE);            // enable keypad
   cbreak();                        // character input
   nodelay(win,true);               // don't wait
   this->wyx();
}
/**
 * Diagnostic to display field associations
 */
void topvhost::diagnose()
{
   Config *conf = this->conf;
   map<int,int>::iterator idb;
   string sep="",s;
   unsigned j;

   cout << "Configuration File: " << conf->path << endl;
   cout << "Active Section:     " << conf->section << endl;
   s = conf->get("hdr_format");
   cout << "Page Header: '" << s << "'" << endl;
   s = conf->get("row_format");
   cout << "Output (Needs): '" << s << "'" << endl;
   for(j=0;j<this->needs.size();j++) {
      cout << sep << j << " = " << this->needs[j];
      sep = ", ";
      }
   cout << endl << "Input (Gets): '" << this->logfmt << "'" << endl;
   sep = "";
   for(j=0;j<this->gets.size();j++) {
      cout << sep << j << " = " << this->gets[j];
      sep = ", ";
      }
   cout << endl << "Transfer Map (Gets => Needs)" << endl;
   sep = "";
   for(idb = this->xfer.begin();idb != this->xfer.end();++idb) {
      cout << sep << idb->first << " => " << idb->second;
      sep = ", ";
      }
   cout << endl << "Domain name @" << this->nidx << endl;
}
/**
 * Shutdown the application
 */
void topvhost::end()
{
   if (this->window) {
      endwin();
      this->window = 0;
      }
}
/**
 * Initialize the application
 */
void topvhost::init(appOptions options)
{
   Config *conf = this->conf = new Config(options.configFile);
   unsigned j,k,nidx=0;

   if (options.configSection)
      conf->load(options.configSection);
   this->options  = options;
   this->logfmt   = conf->get("log_format");
   this->parseSpec(&this->hdr,"hdr_format");
   this->parseSpec(&this->row,"row_format");
   vhost::getNeeds(this);                    // set needs
   vhost::parseApacheLogFormat(this);        // set gets

   for(j = 0;j<this->needs.size();j++) {     // xfer maps gets to needs
      string ch  = this->needs[j];
      int idx = vhost::OUTPUT_FIELDS+j;
      if (ch == "v")                         // domain does not come from log record
         nidx = idx;
      else for(k=0;k<this->gets.size();k++)
         if (ch == this->gets[k]) {
            this->xfer[k] = idx;
            break;
            }
      }
   this->nidx = nidx;
   if (options.diagnose)
      this->diagnose();
}
/**
 * Handle the keyboard
 */
int topvhost::input()
{
   int ch;

   if (this->exit || !this->window) return 'q';
   if (this->resize) {
      this->resize = 0;
      endwin();
      refresh();
      this->wyx();
      }
   switch(ch=getch()) {
      case 'b':   case 'B':   case KEY_PPAGE:
         if (this->spos>0) {
            this->spos -= this->ssz;
            if (this->spos<0) this->spos = 0;
            }
         break;
      case 'd':
         this->debug = !this->debug;
         break;
      case 'f':   case 'F':   case KEY_NPAGE:
         if ((this->spos+this->ssz)<this->total)
            this->spos += this->ssz;
         break;
      case 'h':   case 'H':
         this->order = topvhost::BY_HITS;
         this->paint = true;
         break;
      case 'n':   case 'N':
         this->order = topvhost::BY_NAME;
         this->paint = true;
         break;
      case 'q':   case 'Q':
         ch = 'q';
         break;
      case 'r':   case 'R':
         ch = 'r';
         break;
      case 't':   case 'T':
         this->order = topvhost::BY_ACCESS;
         this->paint = true;
         break;
      }
   return ch;
}
/**
 * Parse formatting specification
 */
void topvhost::parseSpec(std::vector<fSpec>* v,const char *name)
{
   Config *conf            = this->conf;
   string src              = conf->cntrl(conf->get(name));
   string::iterator fmt    = src.begin();
   string s1="-+#0 ";
   string s2=".01234567890";
   string fill,token;
   fSpec spec;
   int ch;

   fill.clear();
   while( fmt != src.end() ) {
      if ((ch = *fmt++)=='%') {
         if (fill.length()>0) {
            spec.field  = ' ';
            spec.format = fill;
            v->push_back(spec);
            fill.clear();
            }
         token = ch;
         if (s1.find(*fmt)!= string::npos)
           token += *fmt++;
         while(s2.find(*fmt)!= string::npos)
           token += *fmt++;
         if (*fmt == '{')
            do { token += ch = *fmt++;} while(fmt != src.end() && ch != '}');
         spec.field  = *fmt++;
         spec.format = token;
         v->push_back(spec);
         }
      else fill += ch;
      }
   if (fill.length()>0) {
      spec.field  = ' ';
      spec.format = fill;
      v->push_back(spec);
      }
}
/**
 * Thin wrapper for curses macro
 */
void topvhost::wmove(int y,int x)
{
   move(y,x);
}
/**
 * Thin wrappers for curses
 */
void topvhost::wprint(string fmt, float val)
{
   char buf[WBUF_LEN];
   fmt += 'g';
   snprintf(buf,WBUF_LEN,fmt.c_str(), val);
   this->wputs(buf);
}
/**
 * Thin wrappers for curses
 */
void topvhost::wprint(string fmt, int val)
{
   char buf[WBUF_LEN];
   fmt += "d";
   snprintf(buf,WBUF_LEN,fmt.c_str(), val);
   this->wputs(buf);
}
/**
 * Thin wrappers for curses
 */
void topvhost::wprint(string fmt, long val)
{
   char buf[WBUF_LEN];
   fmt += "ld";
   snprintf(buf,WBUF_LEN,fmt.c_str(), val);
   this->wputs(buf);
}
/**
 * Thin wrapper for curses macro
 */
void topvhost::wput(int ch)
{
   if (ch=='\n'||this->wx()<(this->maxx-1))
      addch(ch);
}
/**
 * Thin wrapper for curses macro
 */
void topvhost::wputs(const char *str)
{
   while(*str)
      this->wput(*str++);
}
/**
 * Thin wrapper for curses macro
 */
int topvhost::wx()
{
   return getcurx(static_cast<WINDOW *>(this->window));
}
/**
 * Thin wrapper for curses macro
 */
int topvhost::wy()
{
   return getcury(static_cast<WINDOW *>(this->window));
}
/**
 * Get screen size
 */
void topvhost::wyx()
{
   WINDOW*win = static_cast<WINDOW *>(this->window);
   int n=0;

   getmaxyx(win,this->maxy,this->maxx);
   for(unsigned i=0;i<this->hdr.size();i++)
      if (this->hdr[i].field==' ') {
         string s = this->hdr[i].format;
         for (unsigned j=0;j<s.length();j++)
            if (s[j]=='\n') n++;
         }
   if (n==0)
      throw runtime_error("Header requires end of line");
   this->ssz  = (this->maxy - 1 - n)/this->sunit;     // round down!
   this->spos = 0;
}
/**
 * Main loop
 */
int main(int argc, char **argv)
{
   static const char* cmds[] = {
      "d", "init",   "0", "Diagnose Setup",
      "f", "file",   "1", "Configuration file",
      "i", "init",   "0", "Parse logs at startup",
      "s", "select", "1", "Select Configuration",
      "v", "verify", "0", "Verify Configuration",
      "h", "help",   "0", "This help"
      };
   static appOptions options = {0, 0, false, false };
   int nopts = sizeof(cmds)/(4*sizeof(char *));
   struct option* long_options = new option[nopts];
   string hasargs="012";
   string short_options;
   int c,i,j,x0,y0;
   Monitor mon;

   for(i=j=0;i<nopts;i++,j+=4) {
      long_options[i].name      = cmds[j+1];
      long_options[i].has_arg   = hasargs.find(cmds[j+2]);
      long_options[i].flag      = NULL;
      long_options[i].val       = cmds[j][0];
      short_options += cmds[j];
      if (long_options[i].has_arg!=0) short_options += ":";
      }
   do {
      c = getopt_long (argc, argv, short_options.c_str(), long_options, NULL);
      switch(c) {
         case 'd':
            options.diagnose = true;
            break;
         case 'f':
            options.configFile = optarg;
            break;
         case 'i':
            options.init   = true;
            break;
         case 's':
            options.configSection = optarg;
            break;
         case 'h':   case '?':
            cerr << "\nUsage: " << argv[0] << " [options]" << endl << endl;
            cerr << "Monitor Virtual Hosts" << endl << endl;
            cerr << "  " << "Options:" << endl;
            for(i=0;long_options[i].name!=NULL;i++) {
               char buf[128];
               sprintf(buf,"--%-10s, -%c %s %s",
                  long_options[i].name, long_options[i].val,
                  long_options[i].has_arg? "[]":"  ",cmds[i*4+3]);
               cerr << "     " << buf << endl;
               }
            cerr << endl;
            exit(1);
         case 'v':
            options.verify = true;
         }
      } while (c!=-1);

   try {
      app.init(options);
      mon.init(&app);
      while(mon.getInput(&app)) {
         vhost *inst = mon.getHosts(&app);
         mon.paintHeader(&app);
         y0 = app.wy();
         x0 = app.wx();
         printw("\n");
         for(i=app.spos;i<app.total;i++) {
            inst[i].paintHost(&app);
            if (app.wy()>=(app.maxy-1)) break;
            }
         clrtobot();
         move(y0,x0);
         refresh();
         sleep(1);
         }
      mon.quit(&app);
      if (app.exit)
         cerr << " Caught signal " << app.exit << endl;
      }
   catch (exception &e) {
      mon.quit(&app);
      cerr << e.what() << endl;
      }
   return 0;
}
